"""Context is used in parsing."""

import importlib
import types


class Context:
    """Object that gets passed to structs to set runtime context.

    Many of the arrays in the save file have their size set by the value
    of max_players and max_teams. There are also many arrays whose size
    is set by the number of items in the enums generated at runtime from
    the XML files. For example the `unit_created_counts` array is sized
    based on the number of unit types in the `UnitType` enum.
    """
    def __init__(
        self, max_players: int = 19, max_teams: int = 0, ai_survivor: bool = False
    ):
        """Initialize a new Context object.

        Args:
            max_players (int): Civ4 DLL can be compiled with a different
                MAX_PLAYERS value. Default is 19.
            max_teams (int): Optional arg to override default of
                max_teams == max_players.
            ai_survivor (bool): Whether or not to use enums created from
                ai survivor XML files. Defaults to False.
        """
        self.max_players = max_players
        self.max_teams = max_teams if max_teams else max_players
        self.ai_survivor = ai_survivor
        self.enums: types.ModuleType
        if ai_survivor:
            self._module = "ai_survivor"
        else:
            self._module = "vanilla"
        self.enums = importlib.import_module(f"civ4save.enums.{self._module}")

    def __str__(self):
        """Return string representation of the Context."""
        mp = self.max_players
        mt = self.max_teams
        ai = self.ai_survivor
        m = self._module
        return f"Context(max_players={mp}, max_teams={mt}, ai_survivor={ai}, enums={m})"



"""Used in `SaveFile.plots`."""
from typing import List

import attrs

from civ4save.enums import vanilla as e


@attrs.define(slots=True)
class Plot:
    """Represents a plot."""
    x: int
    y: int
    ownership_duration: int
    improvement_duration: int
    starting_plot: bool
    hills: bool
    can_be_worked: bool
    irrigated: bool
    owner: int
    plot_type: e.PlotType
    terrain_type: e.TerrainType
    feature_type: e.FeatureType
    bonus_type: e.BonusType
    improvement_type: e.ImprovementType
    yields: List[int]

    @classmethod
    def from_struct(cls, cv_plot):
        """Return `Plot` from cv_plot parsed struct."""
        plot_type = e.PlotType[cv_plot.plot_type]
        terrain_type = e.TerrainType[cv_plot.terrain_type]
        feature_type = e.FeatureType[cv_plot.feature_type]
        bonus_type = e.BonusType[cv_plot.bonus_type]
        improvement_type = e.ImprovementType[cv_plot.improvement_type]
        return cls(
            cv_plot.x,
            cv_plot.y,
            cv_plot.ownership_duration,
            cv_plot.improvement_duration,
            cv_plot.starting_plot,
            cv_plot.hills,
            cv_plot.potential_city_work,
            cv_plot.irrigated,
            cv_plot.owner,
            plot_type,
            terrain_type,
            feature_type,
            bonus_type,
            improvement_type,
            cv_plot.yields,
        )



"""Responsible for parsing a save file into useful data structures."""
from __future__ import annotations

import zlib
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import construct as C
from lazy_property import LazyProperty

from . import structs, utils
from .objects import Context, GameState, Player, Plot, Settings, get_players


class NotASaveFile(Exception):
    """Raised when file can't be parsed as a save file."""
    pass


class ParseError(Exception):
    """Raised when unexpected parsing error."""
    pass


class ParserState(Enum):
    """Represent current state of the parse."""

    READY = auto()
    METADATA = auto()
    CV_INIT_CORE = auto()
    CV_GAME = auto()
    CV_MAP_BASE = auto()
    CV_MAP_PLOTS = auto()
    PARTIAL_PARSE = auto()
    FULL_PARSE = auto()


class SaveFile:
    """State machine that lazily parses data as needed."""

    def __init__(
        self,
        file: Union[str, Path],
        context: Optional[Context] = None,
        debug: bool = False,
    ):
        """Read and decompress the file, but do not parse anything yet.

        Args:
            file (str | Path): File to be parsed.
            context (Context | None): Optional context overriding default.
            debug (bool): Whether to print detailed debug info.
                Defaults to False.
        """
        self.file = file
        self.context = context if context else Context()
        self.debug = debug

        self._idx: int = 0  # current index in the byte array

        self._z_start: int = 0
        self._z_end: int = 0

        self._raw_bytes: bytes
        self._header_bytes: bytes
        self._compressed_bytes: bytes
        self._uncompressed_bytes: bytes
        self._tail_bytes: bytes
        self._bytes: bytes = self._read()

        self._version: int = 0
        self._metadata: Optional[C.Container[Any]] = None
        self._cv_init_core_data: Optional[C.Container[Any]] = None
        self._cv_game_data: Optional[C.Container[Any]] = None

        self._cv_map_base_data: Optional[C.Container[Any]] = None
        self._plots: List[C.Container[Any]] = []
        self._areas: List[C.Container[Any]] = []

        self._parser_state: ParserState = ParserState.READY

    def _read(self, wbits: int = 15) -> bytes:
        """Read and decompress file.

        Find the index in where the zlib magic header is, then find the end
        of the compressed bytes. Then decompress and return the bytes.
        """
        with open(self.file, "rb") as f:
            data = f.read()
        self._raw_bytes = data

        magic_number = bytes.fromhex("789c")  # default compression
        z_start = data.find(magic_number)
        z_end = self._find_zlib_end(data, z_start)
        if z_start < 0 or z_end < 0:
            raise NotASaveFile("This is not a .CivBeyondSwordSave file")
        self._z_start = z_start
        self._z_end = z_end

        decomp_obj = zlib.decompressobj(wbits=wbits)
        uncompressed_data = decomp_obj.decompress(data[z_start:z_end])

        self._header_bytes = data[:z_start]
        self._compressed_bytes = data[z_start:z_end]
        self._uncompressed_bytes = uncompressed_data
        self._tail_bytes = data[z_end:]

        return data[:z_start] + uncompressed_data + data[z_end:]

    def parse(self) -> ParserState:
        """Only parse what needs to be parsed."""
        while True:
            if self.debug:
                print(self._parser_state)

            if self._parser_state == ParserState.READY:
                self._parser_state = ParserState.METADATA

            elif self._parser_state == ParserState.METADATA:
                self._metadata = self._parse_metadata()

            elif self._parser_state == ParserState.CV_INIT_CORE:
                self._cv_init_core_data = self._parse_cv_init_core()

            elif self._parser_state == ParserState.CV_GAME:
                self._cv_game_data = self._parse_cv_game()

            elif self._parser_state == ParserState.CV_MAP_BASE:
                self._cv_map_base_data = self._parse_cv_map_base()

            elif self._parser_state == ParserState.CV_MAP_PLOTS:
                self._cv_map_plots_data = self._parse_cv_map_plots()

            elif self._parser_state == ParserState.PARTIAL_PARSE:
                break

            elif self._parser_state == ParserState.FULL_PARSE:
                break

        return self._parser_state

    def _parse_metadata(self) -> C.Container[Any]:
        """Parses just the first few bytes of CvInitCore.

        Fail loud and early if the version number is unrecognized.
        """
        struct = structs.metadata()
        data = struct.parse(self._bytes)
        self._idx += data._idx

        if data.version != 302:
            raise NotImplementedError(f"Unkown version={self.version}")
        self._version = data.version
        # Move to next state
        self._parser_state = ParserState.CV_INIT_CORE
        return data

    def _parse_cv_init_core(self) -> C.Container[Any]:
        """Parses CvInitCore."""
        struct = structs.cv_init_core(self.context)
        data = struct.parse(self._bytes[self._idx : self._z_start])
        self._idx += data._idx
        # Move to next state
        self._parser_state = ParserState.CV_GAME
        return data

    def _parse_cv_game(self) -> C.Container[Any]:
        """Parses CvGame."""
        struct = structs.cv_game(self.context)
        data = struct.parse(self._bytes[self._idx :])
        self._idx += data._idx
        # Move to next state
        self._parser_state = ParserState.CV_MAP_BASE
        return data

    def _parse_cv_map_base(self) -> C.Container[Any]:
        """Parses everything in CvMap up to the plots array."""
        struct = structs.cv_map_base(self.context)
        data = struct.parse(self._bytes[self._idx :])
        self._idx += data._idx
        # Move to next state
        self._parser_state = ParserState.PARTIAL_PARSE
        return data

    def _parse_cv_map_plots(
        self, x: Optional[int] = None, y: Optional[int] = None
    ) -> List[C.Container[Any]]:
        """Parses plots checking that at least the x, y coords are correct.

        If self.debug=True print detailed debugging info on failure
        to parse else just stop and return the plots that could be parsed.
        """
        self._start_cv_map_plots = self._idx
        struct = structs.cv_plot(self.context)

        grid_width, grid_height = self.map_size
        if x and y:
            num = utils.calc_plot_index(grid_width, x, y)
        else:
            num = grid_width * grid_height

        next_x, next_y = 0, 0
        for n in range(num):
            try:
                plot = struct.parse(self._bytes[self._idx :])
                assert (plot.x, plot.y) == (next_x, next_y)
            except (C.ConstructError, AssertionError) as ex:
                self._parser_state = ParserState.PARTIAL_PARSE
                if self.debug:
                    self._print_debug_plot(n, ex)
                break
            next_x, next_y = utils.next_plot(plot.x, plot.y, grid_width, grid_height)

            self._idx += plot.end_plot_index
            self._plots.append(plot)

        if len(self._plots) == grid_width * grid_height:
            self._parser_state = ParserState.FULL_PARSE

        return self._plots

    def _print_debug_plot(self, num: int, ex: Exception):
        """Logs the last successfully parsed plot, failing plot, and exception."""
        struct = structs.cv_plot_debug(self.context)
        grid_width, grid_height = self.map_size
        last_plot = self._plots[-1]
        print(f"Parsed {len(self._plots)}/{num} plots")
        print(f"Parsed {self._idx}/{len(self._bytes)} bytes")
        print(f"map_size={grid_width}x{grid_height}")
        print(f"Last good plot: x={last_plot.x} y={last_plot.y}")
        print(last_plot)
        print()
        debug_plot = struct.parse(self._bytes[self._idx :])
        print("Failing Plot")
        print("------------")
        print(debug_plot)
        print(ex)
        self._idx += debug_plot._idx
        print(self._idx)

    @staticmethod
    def _find_zlib_end(data: bytes, z_start: int) -> int:
        """Used to decompress file.

        Binary search to find byte index where incomplete data stream becomes
        incorrect data stream. I'm not sure if the guess_limit makes sense but
        I do want some check to an infinite loop / hang.

        https://forums.civfanatics.com/threads/need-a-little-help-with-editing-civ4-save-files.452707/
        """
        guess_limit = 50

        low, high = z_start, len(data)
        prev = 0
        guesses = 0
        while True:
            z_end = (low + high) // 2
            if z_end == prev:
                return z_end
            if guesses > guess_limit:
                raise NotASaveFile("Could not find zlib end byte index")
            try:
                zlib.decompress(data[z_start:z_end])
            except Exception as ex:
                prev = z_end
                if "incomplete" in ex.__str__():
                    low = z_end + 1
                elif "invalid" in ex.__str__():
                    high = z_end - 1
            guesses += 1
        raise NotASaveFile("Could not find zlib end byte index")

    @property
    def version(self) -> int:
        """Returns the version of the save file, 302 for BTS vanilla."""
        if self._parser_state.value < ParserState.PARTIAL_PARSE.value:
            self.parse()
        return self._version

    @property
    def current_turn(self) -> int:
        """Returns the current turn."""
        if self._parser_state.value < ParserState.PARTIAL_PARSE.value:
            self.parse()
        if not self._cv_init_core_data:
            raise ParseError("current_turn | cv_init_core_data not parsed?")
        return self._cv_init_core_data.game_turn

    @property
    def map_size(self) -> Tuple[int, int]:
        """Returns map grid size (width x height)."""
        if self._parser_state.value < ParserState.FULL_PARSE.value:
            self._parser_state = ParserState.CV_MAP_PLOTS
            self.parse()
        if not self._cv_map_base_data:
            raise ParseError("map_size | cv_map_base_data not parsed?")
        grid_width = self._cv_map_base_data.grid_width
        grid_height = self._cv_map_base_data.grid_height
        return grid_width, grid_height

    @LazyProperty
    def settings(self) -> Settings:
        """Returns the game's settings."""
        if self._parser_state.value < ParserState.PARTIAL_PARSE.value:
            self.parse()
        return Settings.from_struct(
            self._cv_init_core_data, self._cv_game_data, self._cv_map_base_data
        )

    @LazyProperty
    def game_state(self) -> GameState:
        """Return `GameState` object."""
        if self._parser_state.value < ParserState.PARTIAL_PARSE.value:
            self.parse()
        return GameState.from_struct(self._cv_game_data, self._cv_map_base_data)

    @LazyProperty
    def players(self) -> Dict[int, Player]:
        """Return players Dict."""
        if self._parser_state.value < ParserState.PARTIAL_PARSE.value:
            self.parse()
        return get_players(self._cv_init_core_data, self._cv_game_data)

    @LazyProperty
    def plots(self) -> List[Plot]:
        """Return the Plots list."""
        if self._parser_state.value < ParserState.FULL_PARSE.value:
            self._parser_state = ParserState.CV_MAP_PLOTS
            self.parse()
        return [Plot.from_struct(p) for p in self._plots]

    def get_plot(self, x: int, y: int) -> Optional[Plot]:
        """Return `Plot` matching the given coordinates (x, y)."""
        if self._parser_state.value < ParserState.FULL_PARSE.value:
            self._parser_state = ParserState.CV_MAP_PLOTS
            self.parse()
        plot_index = utils.calc_plot_index(self.map_size[0], x, y)
        try:
            return self.plots[plot_index]
        except IndexError:
            return None

    def get_player(self, player_idx: int) -> Optional[Player]:
        """Return `Player` at the given player idx."""
        if self._parser_state.value < ParserState.PARTIAL_PARSE.value:
            self.parse()
        try:
            return self.players[player_idx]
        except KeyError:
            return None

    @property
    def file_layout(self) -> str:
        """Return formatted str showing the files byte layout."""
        compression_ratio = len(self._uncompressed_bytes) / len(self._compressed_bytes)
        strings = [
            f"\traw bytes: {len(self._raw_bytes)}",
            f"\theader bytes (no compression): {len(self._header_bytes)}",
            f"\tzlib compressed bytes: {len(self._compressed_bytes)}",
            f"\tzlib after decompression bytes: {len(self._uncompressed_bytes)}",
            f"\ttail bytes (no compression): {len(self._tail_bytes)}",
            f"\tcompression ratio: {compression_ratio}",
        ]
        return "\n".join(strings)

    def __str__(self) -> str:
        """Return string representation of the `SaveFile`."""
        v = self.version
        sz = len(self._raw_bytes)
        try:
            n = self.file.name  # type: ignore
            return f"SaveFile(file={n}, version={v}, size={sz})"
        except AttributeError:
            f = self.file
            return f"SaveFile(file={f}, version={v}, size={sz})"



"""Some potentially interesting stuff with the Civs."""
import json
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import importlib_resources
import xmltodict

from ..enums.vanilla import (
    BuildingType,
    CivilizationType,
    LeaderHeadType,
    TechType,
    UnitType,
)


@dataclass
class Civ:
    """Object holding Civ attributes."""
    type: CivilizationType
    description: str
    short_description: str
    cities: List[str]
    unique_building: BuildingType
    unique_unit: UnitType
    starting_techs: Tuple[TechType, TechType]
    leaders: List[LeaderHeadType]

    @classmethod
    def from_dict(cls, d: dict) -> "Civ":
        """Create a new Civ object from a dictionary."""
        techs = [TechType[t] for t in d["starting_techs"]]
        starting_techs = techs[0], techs[1]
        return cls(
            type=CivilizationType[d["type"]],
            description=d["description"],
            short_description=d["short_description"],
            cities=d["cities"],
            unique_building=BuildingType[d["unique_building"]],
            unique_unit=UnitType[d["unique_unit"]],
            starting_techs=starting_techs,
            leaders=[LeaderHeadType[ld] for ld in d["leaders"]]
        )


@lru_cache(1)
def _get_civs_from_xml_files(xml_file: str | Path) -> Dict[str, Civ]:
    """Create a mapping of short_description -> Civ.

    Ex. "Mali": Civ(CivilizationType.CIVILIZATION_MALI, ...)
    """
    civs = {}

    text_map_json = (
        importlib_resources.files("civ4save.contrib.data")
        .joinpath("text_map.json")
        .read_text()
    )
    text_map = json.loads(text_map_json)

    with open(xml_file, "r") as f:
        data = xmltodict.parse(f.read())

    civ_infos = data["Civ4CivilizationInfos"]["CivilizationInfos"]["CivilizationInfo"]
    for civ_info in civ_infos:
        type = CivilizationType[civ_info["Type"]]
        description = text_map.get(civ_info["Description"], civ_info["Type"])
        short_description = text_map.get(civ_info["ShortDescription"], "")
        # TODO actually look up the TXT_KEY
        cities = [text_map.get(c, c) for c in civ_info["Cities"]["City"]]
        try:
            unique_building = BuildingType[
                civ_info["Buildings"]["Building"]["BuildingType"]
            ]
            unique_unit = UnitType[civ_info["Units"]["Unit"]["UnitType"]]
        except TypeError:  # MINOR and BARBARIAN civs
            break
        techs = [TechType[t["TechType"]] for t in civ_info["FreeTechs"]["FreeTech"]]
        starting_techs = techs[0], techs[1]
        try:
            # Multiple Leaders
            leaders = [
                LeaderHeadType[ld["LeaderName"]] for ld in civ_info["Leaders"]["Leader"]
            ]
        except TypeError:
            # Single Leader
            leaders = [LeaderHeadType[civ_info["Leaders"]["Leader"]["LeaderName"]]]
        civ = Civ(
            type,
            description,
            short_description,
            cities,
            unique_building,
            unique_unit,
            starting_techs,
            leaders,
        )
        civs[short_description] = civ
    return civs


@lru_cache(1)
def _get_civ_map() -> dict:
    """Load from json file in contrib/data."""
    civs_json = (
        importlib_resources.files("civ4save.contrib.data")
        .joinpath("civs.json")
        .read_text()
    )
    civs = json.loads(civs_json)
    return civs


def get_civs() -> Dict[str, Civ]:
    """Get all Civs."""
    civs = _get_civ_map()
    return {c: Civ.from_dict(civs[c]) for c in civs}


def get_civ(name: str) -> Optional[Civ]:
    """Get a single Civ by name."""
    civs = _get_civ_map()
    civ_dict = civs.get(name, None)
    if civ_dict:
        return Civ.from_dict(civ_dict)
    return None


if __name__ == "__main__":
    # from ..utils import CustomJsonEncoder
    # CIV_INFOS = Path("xml") / "CIV4CivilizationInfos.xml"
    # civs = _get_civs_from_xml_files(CIV_INFOS)
    # print(json.dumps(civs, cls=CustomJsonEncoder, indent=4))
    print(get_civ("Germany"))



"""The object returned by the `--spoilers` option."""
from typing import List

import attrs

from civ4save.enums import vanilla as e


@attrs.define(slots=True)
class GameState:
    """Supposed to represent state a human player shouldn't necessarily know."""
    total_cities: int
    total_population: int
    nukes_exploded: int
    circumnavigated: bool
    nukes_buildable: bool
    best_land_unit: e.UnitType
    winner: int
    victory: e.VictoryType
    state: e.GameStateType
    scores: List[int]
    # holy cities, corp headquarters
    cities_destroyed: List[str]
    great_people_born: List[str]
    # CvMap
    land_plots: int
    owned_plots: int

    @classmethod
    def from_struct(cls, cv_game, cv_map):
        """Create GameState from the cv_game and cv_map parsed structs."""
        best_land_unit = e.UnitType[cv_game.best_land_unit]
        victory = e.VictoryType[cv_game.victory]
        state = e.GameStateType[cv_game.game_state]
        scores = [s for s in cv_game.ai_player_score if s > 0]
        great_people_born = [gp.name for gp in cv_game.great_people_born]

        return cls(
            cv_game.total_cities,
            cv_game.total_population,
            cv_game.nukes_exploded,
            cv_game.circumnavigated,
            cv_game.nukes_valid,
            best_land_unit,
            cv_game.winner,
            victory,
            state,
            scores,
            cv_game.cities_destroyed,
            great_people_born,
            cv_map.land_plots,
            cv_map.owned_plots
        )



"""enums/base.py"""
from enum import Enum


class GameType(Enum):
    GAME_NONE = -1
    GAME_SP_NEW = 0
    GAME_SP_SCENARIO = 1
    GAME_SP_LOAD = 2
    GAME_MP_NEW = 3
    GAME_MP_SCENARIO = 4
    GAME_MP_LOAD = 5
    GAME_HOTSEAT_NEW = 6
    GAME_HOTSEAT_SCENARIO = 7
    GAME_HOTSEAT_LOAD = 8
    GAME_PBEM_NEW = 9
    GAME_PBEM_SCENARIO = 10
    GAME_PBEM_LOAD = 11
    GAME_REPLAY = 12


class GameStateType(Enum):
    GAMESTATE_ON = 0
    GAMESTATE_OVER = 1
    GAMESTATE_EXTENDED = 2


class TradeableItem(Enum):
    TRADE_ITEM_NONE = -1
    TRADE_GOLD = 0
    TRADE_GOLD_PER_TURN = 1
    TRADE_MAPS = 2
    TRADE_VASSAL = 3
    TRADE_SURRENDER = 4
    TRADE_OPEN_BORDERS = 5
    TRADE_DEFENSIVE_PACT = 6
    TRADE_PERMANENT_ALLIANCE = 7
    TRADE_PEACE_TREATY = 8
    TRADE_TECHNOLOGIES = 9
    TRADE_RESOURCES = 10
    TRADE_CITIES = 11
    TRADE_PEACE = 12
    TRADE_WAR = 13
    TRADE_EMBARGO = 14
    TRADE_CIVIC = 15
    TRADE_RELIGION = 16


class ReplayMessageType(Enum):
    MAJOR_EVENT = 0
    CITY_FOUNDED = 1
    PLOT_OWNER_CHANGE = 2


class PlotType(Enum):
    NO_PLOT = -1
    PLOT_PEAK = 0
    PLOT_HILLS = 1
    PLOT_LAND = 2
    PLOT_OCEAN = 3


class MultiplayerOptionType(Enum):
    NO_MPOPTION = -1
    MPOPTION_SIMULTANEOUS_TURNS = 0
    MPOPTION_TAKEOVER_AI = 1
    MPOPTION_SHUFFLE_TEAMS = 2
    MPOPTION_ANONYMOUS = 3
    MPOPTION_TURN_TIMER = 4



"""Misc. classes and functions that are used by other modules."""

import json
import platform
from dataclasses import asdict, is_dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, Iterator, Tuple

import xmltodict


class CustomJsonEncoder(json.JSONEncoder):
    """Enable serializing dataclasses and Enums."""

    def default(self, obj):
        """Override default."""
        if is_dataclass(obj):
            return asdict(obj)
        elif isinstance(obj, Enum):
            return obj.name
        return super().default(obj)


def renderable_filepath(path: Path) -> str:
    """Wrap the path so can be pretty-printed by `rich` even if spaces in path.

    Args:
        path (Path): Path to make renderable.

    Returns:
        str: The string representation of `path` with `repr.path` formatting.
    """
    return f"[repr.path]{str(path)}[/repr.path]"


def calc_plot_index(grid_width: int, x: int, y: int) -> int:
    """Calculate array index of a coordinate pair (x, y) based on grid size.

    Args:
        grid_width (int): Width of the map.
        x (int): X coordinate
        y (int): Y coordinate

    Returns:
        int: The array index of the coordinate.
    """
    index = (grid_width * y) + x
    return index


def next_plot(
    x: int, y: int, grid_width: int, grid_height: int
) -> Tuple[int, int]:
    """Return the next coordinate pair (x, y) taking into account grid size.

    Args:
        x (int): X coordinate
        y (int): Y coordinate
        grid_width (int): Width of the map.
        grid_height (int): Height of the map.

    Returns:
        tuple[int, int]: (x, y) pair of the next plot in array.
    """
    next_x, next_y = x + 1, y

    if x == grid_width - 1:
        next_x = 0
        next_y = y + 1
    return next_x, next_y


def get_enum_length(e: Enum) -> int:
    """Ignore the negative value member of the enum when getting member count.

    Args:
        e (Enum): The enum to calculate number of members of.

    Returns:
        int: Number of members.
    """
    return len([m for m in e.__members__ if e[m].value >= 0])  # type: ignore


def unenumify(name: str) -> str:
    """Transform enum member name to titlized string.

    Args:
        name (str): Enum member name.

    Returns:
        str: Titlized string with underscores replaced by spaces.
    """
    strip_leading = str(name).split("_")[1:]
    return " ".join(strip_leading).title()


def get_game_dir() -> Path:
    """Look in various places for the Civ4 BTS game directory.

    Returns:
        Path: The found game directory.

    Raises:
        FileNotFoundError: If no game directory could be found.
    """
    leaf = "Steam/steamapps/common/Sid Meier's Civilization IV Beyond the Sword"
    possible_locations = [
        Path(r"C:\Program Files (x86)") / leaf,
        Path.home() / ".var/app/com.valvesoftware.Steam/data" / leaf,
        Path.home() / ".local/share" / leaf,
    ]
    for p in possible_locations:
        if p.exists():
            return p
    else:
        raise FileNotFoundError("Could not locate BTS game directory")


def get_saves_dir(sub_dir: str = "single") -> Path:
    """Look in various places for the Civ4 BTS saves directory.

    Returns:
        Path: The found saves directory.

    Raises:
        FileNotFoundError: If no saves directory could be found.
    """
    if platform.system() == "Windows":
        saves_dir = (
            Path.home()
            / "Documents"
            / "My Games"
            / "beyond the sword"
            / "Saves"
            / sub_dir
        )
    else:
        saves_dir = Path.home().joinpath(
            ".local/share/Steam/steamapps/compatdata",
            "8800/pfx/drive_c/users/steamuser",
            "My Documents/My Games/Beyond the Sword/Saves",
            sub_dir,
        )
    if not saves_dir.exists():
        raise FileNotFoundError("Could not locate saves directory.")
    return saves_dir


def get_xml_dir() -> Path:
    """Look in various places for the Civ4 BTS XML directory.

    Returns:
        Path: The found XML directory.

    Raises:
        FileNotFoundError: If no XML directory could be found.
    """
    game_dir = get_game_dir()
    bts_xml = game_dir / "Beyond the Sword" / "Assets" / "XML"
    if not bts_xml.exists():
        raise FileNotFoundError("Could not locate XML directory")
    return bts_xml


def clear_auto_saves() -> None:
    """Delete all autosaves."""
    auto_saves = get_saves_dir() / "single/auto"
    for save in auto_saves.iterdir():
        print(save.name)
        save.unlink()


def get_xml_text_files() -> Iterator[Path]:
    """Iterator of all XML/Text files."""
    game_dir = get_game_dir()
    expansions = ["", "Warlords", "Beyond the Sword"]
    stem = "Assets/XML/Text"
    for expansion in expansions:
        text_dir = game_dir / expansion / stem
        if text_dir.exists():
            for file in text_dir.iterdir():
                yield file


def make_text_map(lang: str = "English") -> Dict[str, str]:
    """Create a Dict mapping TEXT_KEY* to human readable text in language `lang`.

    Works from vanilla to Warlords to BTS so the most recent occurence of a
    TEXT_KEY will be the one in the mapping.

    Args:
        lang (str): Language to map. Defaults to 'English'.

    Returns:
        Dict[str, str]: Dictionary of TEXT_KEY* to text.
    """
    text_map = {}
    for file in get_xml_text_files():
        with open(file, mode="r", encoding="ISO-8859-1") as f:
            data = xmltodict.parse(f.read(), encoding="ISO-8859-1")
        try:
            data["Civ4GameText"]["TEXT"]
        except KeyError:
            continue
        for text in data["Civ4GameText"]["TEXT"]:
            tag, name = text["Tag"], text[lang]
            try:
                new_name = name.get("Text", None)
            except AttributeError:
                text_map[tag] = name
                continue
            text_map[tag] = new_name
    return text_map



"""Provides civ4save command line tool.

Commands:
    parse: Uncompress and parse a .CivBeyondSwordSave file.
    gamefiles: Find and print commonly used game folders.
    leaders: List on or all of the Leaders found in the XML files.
    civs: List one or all of the Civilizations found in the XML files.
    xml: Generate python code or JSON from XML files.

Examples:
```
$ civ4save parse Rome.CivBeyondSwordSave
$ civ4save gamefiles
$ civ4save leaders Shaka
$ civ4save civs Germany
$ civ4save xml --text-map --lang Spanish
```
"""

import json
from functools import partial
from pathlib import Path
from typing import Callable, List, Tuple

import click
from rich import print

from . import __version__, utils
from .contrib.civs import get_civ, get_civs
from .contrib.leaders import get_leader, leader_attributes, rank_leaders
from .objects import Context
from .save_file import SaveFile
from .xml_files import make_enums as write_enums

TEXT_MAP_LANGS = ["English", "French", "German", "Italian", "Spanish"]


@click.group()
@click.version_option(__version__)
def cli():  # noqa: D103
    pass


@cli.command()
@click.option(
    "--max-players",
    default=19,
    type=int,
    help="Needed if you have changed MAX_PLAYERS value in CvDefines.h",
)
@click.option(
    "--settings",
    is_flag=True,
    show_default=True,
    default=False,
    help="Basic settings only. Nothing that would be unknown to the human player",
)
@click.option(
    "--spoilers",
    is_flag=True,
    show_default=True,
    default=False,
    help="Extra info that could give an advantage to human player",
)
@click.option(
    "--player",
    default=-1,
    type=int,
    help="Only show data for a specific player idx. Defaults to the human player",
)
@click.option(
    "--list-players",
    is_flag=True,
    show_default=True,
    default=False,
    help="List all player (idx, name, leader, civ) in the game",
)
@click.option(
    "--ai-survivor",
    is_flag=True,
    show_default=True,
    default=False,
    help="Use XML settings from AI Survivor series",
)
@click.option(
    "--debug",
    is_flag=True,
    show_default=True,
    default=False,
    help="Print detailed debugging info",
)
@click.option(
    "--json",
    "json_",
    is_flag=True,
    show_default=True,
    default=False,
    help="Format output as JSON. Default is text",
)
@click.argument("file", type=click.Path(exists=True, path_type=Path))
def parse(
    max_players,
    settings,
    spoilers,
    player,
    list_players,
    ai_survivor,
    debug,
    json_,
    file,
) -> None:
    """Parse a .CivBeyondSwordSave file.

    FILE is a save file or directory of save files
    """
    context = Context(max_players=max_players, ai_survivor=ai_survivor)
    save = SaveFile(file=file, context=context, debug=debug)
    print(save)
    save.parse()

    print_fn = print
    if json_:
        # call json.dumps on the arg before calling print
        j = partial(json.dumps, indent=4, cls=utils.CustomJsonEncoder)
        print_fn = partial(print, j)

    if spoilers:
        print_fn(save.game_state)
    if player > -1:
        print_fn(save.get_player(player))
    if list_players:
        print_fn(save.players)
    print_fn(save.settings)
    return


@cli.command(help="Find and print relevant game files paths.")
def gamefiles():
    """Print commonly used Civ4 file paths to `stdout`."""
    names_and_getters: List[Tuple[str, Callable[[], Path]]] = [
        ("Game Folder", utils.get_game_dir),
        ("Saves Folder", utils.get_saves_dir),
        ("XML Folder", utils.get_xml_dir),
    ]

    for name, get_dir_func in names_and_getters:
        print(f"[bold]{name}[/bold]")
        print(f"[bold]{'-'*len(name)}[/bold]")
        try:
            dir = get_dir_func()
            print(utils.renderable_filepath(dir))
        except FileNotFoundError:
            pass


@cli.command()
@click.option(
    "--enums", is_flag=True, default=False, help="Transform XML files into Python Enums"
)
@click.option(
    "--text-map",
    is_flag=True,
    default=False,
    help="Create JSON mapping TEXT_KEY to LANG.",
)
@click.option(
    "--lang",
    type=click.Choice(TEXT_MAP_LANGS),
    default="English",
    help="Language to map TEXT_KEY's to",
)
def xml(enums: bool, text_map: bool, lang: str):
    """Generate python code or JSON from the XML files."""
    if enums:
        write_enums()
        return
    if text_map:
        mapping = utils.make_text_map(lang)
        print(json.dumps(mapping, indent=4))
        return


@cli.command()
@click.option(
    "--sort-by",
    type=str,
    required=False,
    default="",
    help="Leader attribute to sort by. Ex. base_peace_weight",
)
@click.option(
    "-r", "--reverse", is_flag=True, default=False, help="Reverse the sort order"
)
@click.option(
    "-l", "--list", "list_", is_flag=True, default=False, help="List all leaders"
)
@click.option(
    "-a", "--attributes", is_flag=True, default=False, help="List all leader attributes"
)
@click.argument("leader_name", type=str, required=False, default="")
def leaders(
    leader_name: str, sort_by: str, reverse: bool, list_: bool, attributes: bool
):
    """Show Leader or list Leaders optionally sorted by attribute.

    LEADER_NAME examples: Shaka, 'Genghis Khan'
    """
    # List leaders alphabetically
    if list_:
        for ld, _ in rank_leaders("description"):
            print(ld)
        return
    # Rank leaders according to attribute passed to sort_by
    if sort_by:
        try:
            for ld, val in rank_leaders(sort_by, reverse):
                print(f"{ld:17} {val}")
        except AttributeError:
            print(f"Leader has no attribute {sort_by}")
        return
    # List all attributes can sort by
    if attributes:
        for attr in leader_attributes():
            print(attr)
        return
    # Otherwise print the passed Leader
    ld = get_leader(leader_name)
    if not ld:
        print(f"Leader name {leader_name} not recognized")
        return
    print(ld)


@cli.command()
@click.option(
    "-l", "--list", "list_", is_flag=True, default=False, help="List all civs"
)
@click.argument("civ_name", type=str, required=False)
def civs(civ_name: str, list_: bool):
    """Show details for a Civ or list all Civs.

    CIV_NAME examples: Germany, 'Holy Rome'
    """
    if list_:
        for civ_desc in get_civs():
            print(civ_desc)
        return
    civ = get_civ(civ_name)
    if not civ:
        print(f"Civilization name {civ} not recognized.")
        return
    print(civ)


if __name__ == "__main__":
    cli()



"""Classes and functions for dealing with players."""

import re
from typing import Dict, List, Union

import attrs

from civ4save.enums import vanilla as e


@attrs.define(slots=True)
class City:
    """Constructed from replay messages."""

    name: str
    x: int
    y: int
    turn_founded: int
    wonders: List[str] = attrs.field(factory=list)


@attrs.define(slots=True)
class Civics:
    """Maps to a players Civics. Initialized with default Civics."""

    government: e.CivicType = e.CivicType.CIVIC_DESPOTISM
    legal: e.CivicType = e.CivicType.CIVIC_BARBARISM
    labor: e.CivicType = e.CivicType.CIVIC_TRIBALISM
    economy: e.CivicType = e.CivicType.CIVIC_DECENTRALIZATION
    religion: e.CivicType = e.CivicType.CIVIC_PAGANISM


@attrs.define(slots=True)
class Trade:
    """Can be an item like TRADE_OPEN_BORDERS or Bonus like BONUS_COW."""

    item: Union[e.TradeableItem, e.BonusType]
    amount: int


@attrs.define(slots=True)
class TradeDeal:
    """Represent a trade deal between players."""

    first_player: int
    second_player: int
    initial_turn: int
    first_trades: List[Trade]
    second_trades: List[Trade]


@attrs.define(slots=True)
class Player:
    """Main class representing a player's state."""

    idx: int
    name: str
    desc: str
    short_desc: str
    adjective: str
    team: int
    handicap: e.HandicapType
    leader: e.LeaderHeadType
    civ: e.CivilizationType
    score: int
    rank: int
    owned_plots: int = 0
    great_people: List[str] = attrs.field(factory=list)
    cities: List[City] = attrs.field(factory=list)
    religion: e.ReligionType = e.ReligionType.NO_RELIGION
    civics: Civics = attrs.field(factory=Civics)
    trades: List[TradeDeal] = attrs.field(factory=list)
    projects: List[str] = attrs.field(factory=list)

    def adopt_civic(self, new_civic: e.CivicType):
        """Correctly set the new civic."""
        v = new_civic.value
        if v < 5:
            self.civics.government = new_civic
        elif v >= 5 and v < 10:
            self.civics.legal = new_civic
        elif v >= 10 and v < 15:
            self.civics.labor = new_civic
        elif v >= 15 and v < 20:
            self.civics.economy = new_civic
        elif v >= 20:
            self.civics.religion = new_civic

    def pop_city(self, city_name: str) -> City:
        """Remove `City` with `city_name` from cities List and return it."""
        for n, city in enumerate(self.cities):
            if city.name == city_name:
                return self.cities.pop(n)
        raise ValueError(f"player {self.idx} has no city named {city_name}")

    def city_from_xy(self, x: int, y: int) -> City:
        """Return city from cities List at given coordinates (x,y)."""
        for city in self.cities:
            if (city.x, city.y) == (x, y):
                return city
        raise ValueError(f"player {self.idx} has no city @ ({x}, {y})")


def _match_empire_to_player(empire: str, players: Dict[int, Player]) -> int:
    """Given a Civ adjective, return the player idx of that Civ.

    For example: 'Indian' -> CivilizationType.CIVILIZATION_INDIA

    Should probably be reworked to pull from leaders.json in `contrib` but
    works for now.
    """
    all_civs = [(en.name, en.value) for en in e.CivilizationType if en.value > -1]
    # Mathc
    empire = empire.upper()[:3]
    for civ, n in all_civs:
        stemmed = civ.split("CIVILIZATION_")[1][:3]
        if stemmed == empire:
            match = e.CivilizationType(n)
            break
    else:
        match = None
    for idx in players:
        if players[idx].civ == match:
            return idx
    else:
        raise ValueError("Could match empire to player")


def _init_players(cv_init, cv_game) -> Dict[int, Player]:
    players = {}
    for p_idx in range(len(cv_init.civs)):
        civ = cv_init.civs[p_idx]
        if civ == "NO_CIVILIZATION":
            continue
        player = Player(
            p_idx,
            name=cv_init.leader_names[p_idx].name,
            desc=cv_init.civ_descriptions[p_idx].description,
            short_desc=cv_init.civ_short_descriptions[p_idx].short_description,
            adjective=cv_init.civ_adjectives[p_idx].adjective,
            team=cv_init.teams[p_idx],
            handicap=e.HandicapType[str(cv_init.handicaps[p_idx])],
            leader=e.LeaderHeadType[str(cv_init.leaders[p_idx])],
            civ=e.CivilizationType[str(civ)],
            score=cv_game.ai_player_score[p_idx],
            rank=cv_game.ai_player_rank[p_idx],
        )
        players[p_idx] = player
    return players


def _set_player_trade_deals(deals, players) -> Dict[int, Player]:
    def process_trades(trades) -> List[Trade]:
        player_trades = []
        for trade in trades:
            item = e.TradeableItem[trade.item]
            amount = 1
            if trade.item in {"TRADE_GOLD", "TRADE_GOLD_PER_TURN"}:
                amount = trade.extra_data
            elif trade.item == "TRADE_RESOURCES":
                item = e.BonusType(trade.extra_data)  # type: ignore
            player_trades.append(Trade(item, amount))
        return player_trades

    for deal in deals:
        trade_deal = TradeDeal(
            deal.first_player,
            deal.second_player,
            deal.initial_game_turn,
            process_trades(deal.first_trades),
            process_trades(deal.second_trades),
        )
        players[deal.first_player].trades.append(trade_deal)
    return players


def _set_player_data(replay_messages, players) -> Dict[int, Player]:
    """Set from replay messages.

    Read the replay messages and assign ownership of plots, cities, wonders
    as well as player religion and civics
    """
    # local vars for tracking who currently owns cities and plots
    cities: Dict[str, int] = {}
    plots: Dict[str, int] = {}

    def text_to_enum_name(text: str):
        rm_leading = re.sub(".*color=[0-9]+,[0-9]+,[0-9]+,[0-9]+>", "", text)
        rm_trailing = rm_leading.replace("</color>!", "")
        return rm_trailing.replace(" ", "_").upper()

    for msg in replay_messages:
        p_idx = msg.player

        if msg.type == "CITY_FOUNDED":
            city_name = msg.text.split(" is founded")[0]
            city = City(city_name, msg.plot_x, msg.plot_y, msg.turn)
            players[p_idx].cities.append(city)
            cities[city_name] = p_idx

        elif msg.type == "MAJOR_EVENT":
            if "captured" in msg.text:
                city_and_owner = msg.text.split(" was captured")[0]
                city_name = city_and_owner.split("(")[0].strip()
                prev_owner = cities[city_name]
                city = players[prev_owner].pop_city(city_name)
                players[p_idx].cities.append(city)
                cities[city_name] = p_idx
            elif "razed" in msg.text:
                city_name = msg.text.split(" is razed")[0]
                prev_owner = cities[city_name]
                players[prev_owner].pop_city(city_name)
                del cities[city_name]
            elif "completes" in msg.text:
                wonder = msg.text.split(" completes ")[-1]
                if (msg.plot_x, msg.plot_y) == (-1, -1):  # global wonder (ie apollo)
                    players[p_idx].projects.append(wonder)
                    continue
                city = players[p_idx].city_from_xy(msg.plot_x, msg.plot_y)
                city.wonders.append(wonder)
            elif "born" in msg.text:
                gp = msg.text.split(" has been born")[0]
                players[p_idx].great_people.append(gp)
            elif "converts" in msg.text:
                relig_text = text_to_enum_name(msg.text)
                religion = e.ReligionType[f"RELIGION_{relig_text}"]
                players[p_idx].religion = religion
            elif "adopts" in msg.text:
                civic_text = text_to_enum_name(msg.text)
                civic = e.CivicType[f"CIVIC_{civic_text}"]
                players[p_idx].adopt_civic(civic)

            elif "revolts" in msg.text:
                city_name = msg.text.split(" revolts")[0]
                prev_owner = cities[city_name]
                city = players[prev_owner].pop_city(city_name)
                empire = msg.text.split("joins the ")[1].replace(" Empire!", "")
                p_idx = _match_empire_to_player(empire, players)
                players[p_idx].cities.append(city)
            else:
                # print(msg.player, msg.text)
                pass

        elif msg.type == "PLOT_OWNER_CHANGE":
            plot_key = f"{msg.plot_x}-{msg.plot_y}"
            try:
                prev_owner = plots[plot_key]
            except KeyError:
                # First assignment
                plots[plot_key] = p_idx
                players[p_idx].owned_plots += 1
                continue
            # re-assignment
            if p_idx > -1:
                players[p_idx].owned_plots += 1
                if prev_owner > -1:
                    players[prev_owner].owned_plots -= 1
            # owner lost plot but no new owner
            elif p_idx == -1:
                players[prev_owner].owned_plots -= 1
            plots[plot_key] = p_idx

    return players


def _fix_potential_duplicate_cities(players) -> Dict[int, Player]:
    """Hacky fix but works.

    If capital founded but then map regenerated on turn 0 | 1 there will be
    duplicate capital founding messages for the human player.
    """
    for p_idx in players:
        cities = players[p_idx].cities

        new_start_idx = 0
        names = set()
        for n, city in enumerate(cities):
            if city.name in names:
                new_start_idx = n
            else:
                names.add(city.name)
        players[p_idx].cities = cities[new_start_idx:]
    return players


def get_players(cv_init, cv_game) -> Dict[int, Player]:
    """Return players dict with all values set."""
    players = _init_players(cv_init, cv_game)
    players = _set_player_data(cv_game.replay_messages, players)
    players = _set_player_trade_deals(cv_game.deals, players)
    players = _fix_potential_duplicate_cities(players)
    return players
